from logger_local.LoggerComponentEnum import LoggerComponentEnum

MESSAGE_LOCAL_PYTHON_COMPONENT_ID = 259
MESSAGE_LOCAL_PYTHON_COMPONENT_COMPONENT_NAME = "message local"
DEVELOPER_EMAIL = 'akiva.s@circ.zone'
object_message = {
    'component_id': MESSAGE_LOCAL_PYTHON_COMPONENT_ID,
    'component_name': MESSAGE_LOCAL_PYTHON_COMPONENT_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': DEVELOPER_EMAIL
}

DEFAULT_HEADERS = {"Content-Type": "application/json"}

# TEST_API_TYPE_ID = 4

# SMS
SMS_MESSAGE_LENGTH = 160
UNICODE_SMS_MESSAGE_LENGTH = 70



from database_mysql_local.generic_crud_ml import GenericCRUDML
from logger_local.MetaLogger import MetaLogger

from .MessageConstants import object_message
from .Recipient import Recipient

cache = {}


class MessageTemplates(GenericCRUDML, metaclass=MetaLogger, object=object_message):
    def __init__(self):
        super().__init__(default_schema_name="field", default_table_name="field_table")

    def get_potentials_recipients(self, criteria_json: dict, recipients: list[Recipient] = None) -> list[dict]:
        where = self.get_where_by_criteria_json(criteria_json)
        if recipients:
            profile_ids_str = ",".join(str(recipient.get_profile_id()) for recipient in recipients)
            where += f" AND user.profile_id IN ({profile_ids_str})"
        query_for_potentials_recipients = f"""
SELECT DISTINCT user_id, person_id, user_main_email_address, user.profile_id AS profile_id, 
   profile_phone_full_number_normalized, profile_preferred_lang_code
 FROM user.user_general_view AS user
    JOIN group_profile.group_profile_view AS group_profile on group_profile.profile_id = user.profile_id
  WHERE {where} 
"""
        columns = ("user_id, person_id, user_main_email_address, profile_id,"
                   "profile_phone_full_number_normalized, profile_preferred_lang_code")
        self.cursor.execute(query_for_potentials_recipients)
        return [self.convert_to_dict(row, columns) for row in self.cursor.fetchall()]

    def get_where_by_criteria_json(self, criteria_json: dict) -> str:
        # TODO add support to user_external_id in criteria_json
        min_age = criteria_json.get("min_age")
        max_age = criteria_json.get("max_age")
        gender_list_id = criteria_json.get("gender_list_id")
        group_list_id = criteria_json.get("group_list_id")
        self.logger.info(object={"min_age": min_age, "max_age": max_age, "gender_list_id": gender_list_id,
                                 "group_list_id": group_list_id})
        # profile_id didn't receive messages from this campaign for campaign.minimal_days
        where = "TRUE "
        if min_age is not None:
            where += f" AND TIMESTAMPDIFF(YEAR, person_birthday_date, CURDATE()) >= {min_age}"
        if max_age is not None:
            where += f" AND TIMESTAMPDIFF(YEAR, person_birthday_date, CURDATE()) <= {max_age}"

        if gender_list_id is not None:
            gender_cache_key = ("gender", gender_list_id)
            if gender_cache_key not in cache:
                profile_gender_id_list = self.sql_in_list_by_entity_list_id(
                    schema_name="gender", entity_name="gender", entity_list_id=gender_list_id)
                cache[gender_cache_key] = profile_gender_id_list
            else:
                profile_gender_id_list = cache[gender_cache_key]

            where += " AND profile_gender_id " + profile_gender_id_list

        if group_list_id is not None:
            group_cache_key = ("group", group_list_id)
            if group_cache_key not in cache:
                group_id_list = self.sql_in_list_by_entity_list_id(
                    schema_name="group", entity_name="group", entity_list_id=group_list_id)
                cache[group_cache_key] = group_id_list
            else:
                group_id_list = cache[group_cache_key]
            where += " AND group_profile.group_id " + group_id_list
        return where

    def get_critiria_json(self, message_template_id: int) -> dict:
        if message_template_id in cache:
            return cache[message_template_id]
        query = """
        SELECT DISTINCT min_age, max_age, gender_list_id, group_list_id
          FROM message_template.message_template_view
            JOIN message_template.message_template_message_template_text_block_view AS message_template_message_template_text_block
                ON message_template_message_template_text_block.message_template_id = message_template_view.message_template_id
            JOIN message_template.message_template_text_block_view AS message_template_text_block
                ON message_template_text_block.message_template_text_block_id = message_template_message_template_text_block.message_template_id
            JOIN criteria.criteria_view AS criteria
                ON criteria.criteria_id = message_template_text_block.criteria_id
          WHERE message_template_view.message_template_id = %s
          LIMIT 1  -- TODO: remove
          """  # noqa
        self.cursor.execute(query, (message_template_id,))

        columns = "min_age, max_age, gender_list_id, group_list_id"
        critiria_json = self.convert_to_dict(self.cursor.fetchone(), columns)
        cache[message_template_id] = critiria_json
        return critiria_json

    def get_textblocks_and_attributes_by_form_id(self, form_id: int) -> list[dict]:
        # TODO: get unique (field_id, question_id) from field view for a given field_id

        query = """SELECT form_id AS formId, form_name AS formName, form_message_seq AS formMessageSeq,
        form_message_template_id AS formMessageTemplateId, message_template_id AS messageTemplateId,
        message_template_name AS messageTemplateName, message_template_text_block_seq AS messageTemplateTextBlockSeq,
        message_template_text_block_id AS messageTemplateTextBlockId, question_id AS questionId,
        question_title AS questionTitle, schema_attributes AS schemaAttributes, uischema_attributes AS uiSchemaAttributes,
        question_type_id AS questionTypeId, question_type_name AS questionTypeName, variable_name AS variableName,
        variable_ml_title AS variableMlTitle, field_name AS fieldName, possible_answer AS possibleAnswer
        FROM form.form_general_view WHERE form_id = %s"""
        columns = ("formId", "formName", "formMessageSeq", "formMessageTemplateId", "messageTemplateId",
                   "messageTemplateName", "messageTemplateTextBlockSeq", "messageTemplateTextBlockId",
                   "questionId", "questionTitle", "schemaAttributes", "uiSchemaAttributes", "questionTypeId",
                   "questionTypeName", "variableName", "variableMlTitle", "fieldName", "possibleAnswer")
        self.cursor.execute(query, (form_id,))
        return [self.convert_to_dict(row, ",".join(columns)) for row in self.cursor.fetchall()]

    def get_textblocks_and_attributes(self) -> list[dict]:
        """Returns the message template text block and attributes for the given message template id and profile id.
        The attributes are:
            - sms_body_template
            - email_subject_template
            - email_body_html_template
            - whatsapp_body_template
            ... TODO: add more attributes here

        The attributes are returned as a list of dictionaries, one dictionary
            for each text block in the message template.
        """

        if "get_textblocks_and_attributes" in cache:
            return cache["get_textblocks_and_attributes"]
        textblocks_and_attributes = []
        query = """
SELECT sms_body_template, email_subject_template, email_body_html_template, whatsapp_body_template,
        default_subject_template, default_body_template,
        question.question_id AS questionId, question.question_type_id AS questionTypeId,
        question.where_statement AS whereStatement, question.variable_id AS variableId, question.table_id AS tableId, 
        question.is_visible AS isVisible, question.is_required AS isRequired, question.label_question_id AS labelQuestionId,
        question.default_value AS defaultValue, question.answer_expiration_days AS answerExpirationDays,
        question_ml.title AS questionTitle, question_type.name AS questionTypeName,
        question_type.schema_attributes AS `schemaDefault`, question_type.uischema_attributes AS uiSchemaDefault,
        question.schema_attributes AS `schemaAttributes`, question.uischema_attributes AS uiSchemaAttributes,

        question_ml.lang_code AS langCode, question.hint AS hint,
        message_template_text_block.message_template_text_block_id AS messageTemplateTextBlockId,
        message_template_text_block_type_ml.title AS messageTemplateBlockTypeName,
        message_template_text_block_type_ml.message_template_text_block_type_id AS blockTypeId,
        message_template_message_template_text_block.message_template_id AS messageTemplateId
        -- answer_expiration_days, criteria_id, label_question_id, message_template_text_block_ml_id,

FROM message_template.message_template_text_block_view AS message_template_text_block
    JOIN message_template.message_template_text_block_ml_view AS message_template_text_block_ml
        ON message_template_text_block_ml.message_template_text_block_id = message_template_text_block.message_template_text_block_id
            -- we filter that later per recipient
            -- AND message_template_text_block_ml.lang_code = %s

    JOIN message_template.message_template_message_template_text_block_view AS message_template_message_template_text_block
        ON message_template_message_template_text_block.message_template_text_block_id = message_template_text_block.message_template_text_block_id

    JOIN message_template.message_template_text_block_type_ml_view AS message_template_text_block_type_ml
        ON message_template_text_block_type_ml.message_template_text_block_type_id = message_template_text_block.message_template_text_block_type_id
            AND message_template_text_block_type_ml.lang_code = message_template_text_block_ml.lang_code

    LEFT JOIN question.question_table AS question
        ON question.question_id = message_template_text_block.question_id

    LEFT JOIN question.question_ml_view AS question_ml
        ON question_ml.question_id = question.question_id
            AND question_ml.lang_code = message_template_text_block_ml.lang_code

    LEFT JOIN question.question_type_view AS question_type
        ON question_type.question_type_id = question.question_type_id
"""  # noqa

        self.cursor.execute(query)
        columns = ("sms_body_template", "email_subject_template", "email_body_html_template", "whatsapp_body_template",
                   "default_subject_template", "default_body_template",
                   "questionId", "questionTypeId", "whereStatement", "variableId", "tableId", "isVisible",
                   "isRequired", "labelQuestionId", "defaultValue", "answerExpirationDays",
                   "questionTitle", "questionTypeName",
                   "schemaDefault", "uiSchemaDefault", "schemaAttributes", "uiSchemaAttributes",
                   "langCode", "hint",
                   "messageTemplateTextBlockId", "messageTemplateBlockTypeName", "blockTypeId", "messageTemplateId")
        for inner_row in self.cursor.fetchall():
            text_block_dict = self.convert_to_dict(inner_row, ", ".join(columns))

            text_block_dict["possibleAnswers"] = self._get_possible_answers(
                question_id=text_block_dict["questionId"])
            self.logger.info(object={"text_block_dict": text_block_dict})
            textblocks_and_attributes.append(text_block_dict)
        cache["get_textblocks_and_attributes"] = textblocks_and_attributes

        return textblocks_and_attributes

    def _get_possible_answers(self, question_id: int) -> list[dict]:
        # TODO: get cities etc and insert as possible answer.
        query = "SELECT value FROM question.question_possible_answers_ml_view WHERE question_id = %s"
        self.cursor.execute(query, (question_id,))
        # We will change action in the future.
        return [{"answerValue": row[0], "action": None} for row in self.cursor.fetchall()]



from database_mysql_local.generic_crud import GenericCRUD
from logger_local.MetaLogger import MetaLogger

from .MessageConstants import object_message


# MessageLocal init is too heavy
class GetTestIds(GenericCRUD, metaclass=MetaLogger, object=object_message):
    def __init__(self):
        super().__init__(default_schema_name="message", default_table_name="message_table",
                         default_view_table_name="message_outbox_view", default_id_column_name="message_id")

    def insert_message(self, **kwargs) -> int:
        """Inserts a message into the database"""
        data_json = {}
        for key, value in kwargs.items():
            if value is not None:
                data_json[key] = value

        message_id = self.insert(schema_name="message", data_json=data_json)
        return message_id

    def get_test_message_id(self) -> int:
        return self.get_test_entity_id(entity_name="message",
                                       insert_function=self.insert_message)



from item_local.item import Item

from .MessageLocal import MessageLocal


class SmsMessageLocal(Item, MessageLocal):
    pass  # TODO implement

    def check_message(self):
        # TODO Check that there is only body without subject
        # TODO Check that there is version of plain text without HTML
        # TODO Check the length of the self.__body_after_text_template is in the right length
        pass



import http
import time
from datetime import datetime
from typing import List

from api_management_local.API_Mangement_Manager import APIMangementManager
from api_management_local.Exception_API import (ApiTypeDisabledException,
                                                ApiTypeIsNotExistException,
                                                PassedTheHardLimitException)
from api_management_local.api_limit_status import APILimitStatus
from api_management_local.direct import Direct
from api_management_local.indirect import InDirect
from item_local.item import Item
from logger_local.MetaLogger import MetaLogger
from python_sdk_remote.utilities import our_get_env
from star_local.exception_star import NotEnoughStarsForActivityException

from .ChannelProviderConstants import (AWS_SES_EMAIL_MESSAGE_PROVIDER_ID,
                                       AWS_SNS_SMS_MESSAGE_PROVIDER_ID,
                                       INFORU_MESSAGE_PROVIDER_ID)
from .CompoundMessage import CompoundMessage
from .MessageChannels import MessageChannel
from .MessageConstants import (DEFAULT_HEADERS, SMS_MESSAGE_LENGTH,
                               object_message)
from .MessageImportance import MessageImportance
from .Recipient import Recipient

DEFAULT_SENDER_PROFILE_ID = our_get_env("DEFAULT_SENDER_PROFILE_ID")


class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_message):
    """Message Local Class"""

    def __init__(self, *, recipients: List[Recipient], api_type_id: int = None, campaign_id: int = None,
                 message_template_id: int = None,
                 original_body: str = None, original_subject: str = None,
                 importance: MessageImportance = MessageImportance.MEDIUM, message_id: int = None,
                 is_http_api: bool = None, endpoint: str = None,
                 headers: dict = DEFAULT_HEADERS, external_user_id: int = None, form_id: int = None,
                 sender_profile_id: int = DEFAULT_SENDER_PROFILE_ID, is_test_data: bool = False, **kwargs) -> None:
        """message_dict contains a list of dicts, each with the following keys:
        ["sms_body_template", "email_subject_template", "email_body_html_template",
        "whatsapp_body_template", "question_id", "question_type_id", "question_title", "question_type_name"]"""
        # TODO We should add all fields from message schema in the database
        # (i.e. message_id, scheduled_sent_timestamp, message_sent_status : MessageSentStatus  ...)

        CompoundMessage.__init__(self, message_id=message_id, campaign_id=campaign_id,
                                 message_template_id=message_template_id, body=original_body, form_id=form_id,
                                 subject=original_subject, recipients=recipients, is_test_data=is_test_data)
        # GenericCRUD.__init__(self, default_schema_name="message", default_table_name="message_table",
        #                      default_view_table_name="message_outbox_view", default_id_column_name="message_id",
        #                      is_test_data=is_test_data)

        # init instance variables
        self.message_id = message_id
        self.__original_body = original_body
        self.__original_subject = original_subject
        self.importance = importance
        self._is_http_api = is_http_api
        self._api_type_id = api_type_id
        self._endpoint = endpoint
        self._headers = headers
        self.__external_user_id = external_user_id
        self.__recipients = recipients
        self._campaign_id = campaign_id
        self.__body_after_html_template = None
        self.__subject_after_html_template = None  # TODO: implement
        self.sender_profile_id = sender_profile_id

        # init instances
        self.__indirect = InDirect()
        self.__direct = Direct()
        self.api_management_manager = APIMangementManager()

        # set body after text template, this most be done before calling get_msg_type
        self.__ordered_message_channels = [message_channel for message_channel in MessageChannel]

        # decide how to send (channel & provider) for each recipient
        self.channel_ids_per_recipient = {recipient.get_profile_id(): self.get_msg_type(recipient)
                                          for recipient in recipients}
        self.provider_id_per_recipient = {recipient.get_profile_id(): self.get_message_provider_id(
            message_channel_id=self.channel_ids_per_recipient[recipient.get_profile_id()], recipient=recipient)
            for recipient in recipients}
        # self.message_template_dict_by_campaign_id = self.get_message_template_dict_by_campaign_id(campaign_id=campaign_id)

        # Old code I might need later:
        # if self._campaign_id:
        #     body = self.message_template_dict_by_campaign_id.get(
        #         self._campaign_id).get(recipient.get_preferred_language())
        #     body = self.message_template.get_message_template_textblock_and_attributes(
        #         message_template_id=body, destination_profile_id=recipient.get_profile_id())

    def get_id(self):
        pass

    def get_msg_type(self, recipient: Recipient) -> MessageChannel:
        # TODO: return msg_type (sms, email, whatsapp) based on hours, provider availability, msg length, etc.
        """TODO: make sure we can access:
        1. size of message
        2. message contains html or not
        3. country of recipient
        4. time of the day
        5. preferences of the recipient
        6. attachments type and size 7. cost of sending the message"""
        for message_channel in self.__ordered_message_channels:
            if recipient.get_email_address() is not None:
                return MessageChannel.EMAIL
            elif len(self.get_body_after_text_template(recipient, message_channel)) < SMS_MESSAGE_LENGTH:
                return MessageChannel.SMS
            else:
                return MessageChannel.WHATSAPP

    def get_message_channel(self, recipient: Recipient) -> MessageChannel:
        if recipient.get_profile_id() in self.channel_ids_per_recipient:
            return self.channel_ids_per_recipient[recipient.get_profile_id()]
        else:
            return self.get_msg_type(recipient)

    @staticmethod
    def get_message_provider_id(message_channel_id: MessageChannel, recipient: Recipient) -> int:
        """return message provider"""
        # TODO: implement the logic for the provider_id (remove +972, etc.)
        if message_channel_id == MessageChannel.SMS and recipient.get_normizalied_phone().startswith("+972"):
            provider_id = AWS_SNS_SMS_MESSAGE_PROVIDER_ID
        elif message_channel_id == MessageChannel.EMAIL:
            provider_id = AWS_SES_EMAIL_MESSAGE_PROVIDER_ID
        elif message_channel_id == MessageChannel.WHATSAPP:
            # TODO: or vonage
            provider_id = INFORU_MESSAGE_PROVIDER_ID
        else:
            raise Exception("Can't determine the Message Provider for message_channel_id=" + str(message_channel_id))

        return provider_id

    #TODO get_body_json_after_text_template()
    def get_body_after_text_template(self, recipient: Recipient,
                                     message_channel: MessageChannel = None) -> str:
        # TODO: is this the right way to do it?
        if message_channel is None:
            message_channel = self.get_message_channel(recipient)
            self.logger.debug("message_channel: " + str(message_channel))

        body = ""
        for block in self.get_compound_message_dict(message_channel)[message_channel.name]["bodyBlocks"]:
            for profile_block in block["profileBlocks"]:
                if profile_block["profileId"] == recipient.get_profile_id():
                    body += profile_block["processedTemplate"] + "\n"
        return body

    def _get_body_after_html_template(self) -> str:
        return self.__body_after_html_template

    # def _get_number_of_attachment(self) -> int:
    #     return 0

    def get_subject_after_html_template(self) -> str:
        # Unresolved attribute reference '__subject_after_html_template' for class 'MessageLocal'
        return self.__subject_after_html_template

    # def _get_type_of_attachments(self):
    #     return None

    # api_data To know if to API calls are actually the same and do caching.
    # TODO sender_profile_id = None meaning we let the system to decide who is the sender_profile_id?
    # TODO Do we need user_external as parameter?
    def can_send(self, sender_profile_id: int = None, api_data: dict = None, outgoing_body: dict = None) -> bool:
        sender_profile_id = sender_profile_id or self.sender_profile_id
        if self._is_http_api:
            return self.__can_send_direct(sender_profile_id, api_data=api_data)
        else:
            return self.__can_send_indirect(sender_profile_id, outgoing_body=outgoing_body)

    # api_data To know if to API calls are actually the same and do caching.
    def __can_send_direct(self, sender_profile_id: int, api_data: dict = None) -> bool:
        # TODO: implement sender_profile_id logic
        try:
            # TODO: change try_to_call_api typing
            try_to_call_api_result = self.__direct.try_to_call_api(
                external_user_id=self.__external_user_id,
                api_type_id=self._api_type_id,
                endpoint=self._endpoint,
                outgoing_body=api_data,  # data
                outgoing_header=self._headers
            )
            x = try_to_call_api_result['status_code']
            if x != http.HTTPStatus.OK:
                raise Exception(try_to_call_api_result['text'])
            else:
                return True
        except PassedTheHardLimitException:
            sleep_after_passing_the_hard_limit = self.api_management_manager.seconds_to_sleep_after_passing_the_hard_limit(
                api_type_id=self._api_type_id)
            if sleep_after_passing_the_hard_limit > 0:
                self.logger.info("sleeping for sleep_after_passing_the_hard_limit=" + str(
                    sleep_after_passing_the_hard_limit) + " seconds",
                                 {'sleep_after_passing_the_hard_limit': sleep_after_passing_the_hard_limit})
                time.sleep(sleep_after_passing_the_hard_limit)
            else:
                self.logger.info("No sleeping needed : x= " + str(sleep_after_passing_the_hard_limit) + " seconds")
        except NotEnoughStarsForActivityException:
            self.logger.warning("Not Enough Stars For Activity Exception")

        except ApiTypeDisabledException:
            self.logger.error("Api Type Disabled Exception")

        except ApiTypeIsNotExistException:
            self.logger.error("Api Type Is Not Exist Exception")

        except Exception as exception:
            self.logger.exception(object=exception)
        return False

    def __can_send_indirect(self, sender_profile_id: int = None, outgoing_body: dict = None) -> bool:
        # TODO: implement sender_profile_id logic
        http_status_code = None
        try:
            api_check, self.__api_call_id, arr = self.__indirect.before_call_api(
                external_user_id=self.__external_user_id, api_type_id=self._api_type_id,
                endpoint=self._endpoint,
                outgoing_header=self._headers,
                outgoing_body=outgoing_body
            )
            if arr is None:
                self.__used_cache = False
                if api_check == APILimitStatus.BETWEEN_SOFT_LIMIT_AND_HARD_LIMIT:
                    self.logger.warn("You passed the soft limit")
                if api_check != APILimitStatus.GREATER_THAN_HARD_LIMIT:
                    try:
                        http_status_code = http.HTTPStatus.OK.value
                    except Exception as exception:
                        self.logger.exception(object=exception)
                        http_status_code = http.HTTPStatus.BAD_REQUEST.value
                else:
                    self.logger.info("You passed the hard limit")
                    x = self.api_management_manager.seconds_to_sleep_after_passing_the_hard_limit(
                        api_type_id=self._api_type_id)
                    if x > 0:
                        self.logger.info("sleeping : " + str(x) + " seconds")
                        time.sleep(x)
                        # raise PassedTheHardLimitException

                    else:
                        self.logger.info("No sleeping needed : x= " + str(x) + " seconds")
            else:
                self.__used_cache = True
                self.logger.info("result from cache")
                # print(arr)
                http_status_code = http.HTTPStatus.OK.value
        except ApiTypeDisabledException:
            self.logger.error("Api Type Disabled Exception")

        except ApiTypeIsNotExistException:
            self.logger.error("Api Type Is Not Exist Exception")
        self.logger.info("http_status_code: " + str(http_status_code))
        return http_status_code == http.HTTPStatus.OK.value

    def send(self, body: str = None, compound_message_dict: dict = None,
             recipients: List[Recipient] = None, cc: List[Recipient] = None, bcc: List[Recipient] = None,
             scheduled_timestamp_start: datetime = None,
             scheduled_timestamp_end: datetime = None, **kwargs) -> list[int]:
        pass  # this is an abstract method, but we don't want to make this class abstract

    def after_send_attempt(self, sender_profile_id: int = None, outgoing_body: dict = None,
                           incoming_message: str = None, http_status_code: int = None,
                           response_body: str = None) -> None:
        # TODO: implement sender_profile_id logic
        sender_profile_id = sender_profile_id or self.sender_profile_id
        if self._is_http_api:
            self.__after_direct_send()
        else:
            self.__after_indirect_send(outgoing_body=outgoing_body,
                                       incoming_message=incoming_message,
                                       http_status_code=http_status_code,
                                       response_body=response_body)

    def display(self):
        print("MessageLocal: " + str(self.__dict__))

    def __after_indirect_send(self, outgoing_body: dict, incoming_message: str,
                              http_status_code: int, response_body: str):

        self.__indirect.after_call_api(external_user_id=self.__external_user_id,
                                       api_type_id=self._api_type_id,
                                       endpoint=self._endpoint,
                                       outgoing_header=self._headers,
                                       outgoing_body=outgoing_body,
                                       incoming_message=incoming_message,
                                       http_status_code=http_status_code,
                                       response_body=response_body,
                                       api_call_id=self.__api_call_id,
                                       used_cache=self.__used_cache)

    def __after_direct_send(self):
        pass

    def get_importance(self) -> MessageImportance:
        """get method"""
        return self.importance

    def get_recipients(self) -> List[Recipient]:
        return self.__recipients



import json
from enum import Enum, auto

from phones_local.phones_local import PhonesLocal
from variable_local.variable import VariablesLocal


# TODO Each of them has field_id value in field.field_table, shall we use it?
#   (I didn't find it there)
class ReferenceType(Enum):
    PERSON_ID = auto()
    CONTACT_ID = auto()
    USER_ID = auto()
    PROFILE_ID = auto()


class RecipientType(Enum):
    TELEPHONE_NUMBER = auto()
    EMAIL_ADDRESS = auto()

    UNKNOWN = auto()


class Recipient:
    def __init__(self, contact_id: int = None, person_id: int = None, user_id: int = None, profile_id: int = None,
                 telephone_number: str = None, email_address_str: str = None, preferred_lang_code_str: str = None,
                 first_name: str = None) -> None:
        self.__person_id = person_id
        self.__email_address_str = email_address_str
        self.__contact_id = contact_id
        self.__user_id = user_id
        self.__profile_id = profile_id
        self.__telephone_number = telephone_number
        self.__preferred_lang_code_str = preferred_lang_code_str
        self.__first_name = first_name
        for key, value in self.to_json().items():
            if not key.endswith("id") and key.upper() in RecipientType.__members__:
                self._recipient_type = RecipientType[key.upper()]  # remove the first underscore

        self.__variable_local: VariablesLocal | None = None  # To improve performance, we will get it only when needed.

    def get_profile_variables(self) -> VariablesLocal:
        if self.__variable_local is not None:
            return self.__variable_local
        self.__variable_local = VariablesLocal()
        # TODO: make sure those are stored on the effective_profile_id (using global user_context or sending the user_context).
        self.__variable_local.add(self.__contact_id, "contact_id")
        self.__variable_local.add(self.__person_id, "person_id")
        self.__variable_local.add(self.__user_id, "user_id")
        self.__variable_local.add(self.__profile_id, "profile_id")
        self.__variable_local.add(self.__telephone_number, "telephone_number")
        self.__variable_local.add(self.__email_address_str, "email_address")
        return self.__variable_local

    def get_person_id(self) -> int:
        return self.__person_id

    def get_profile_id(self) -> int:
        return self.__profile_id

    def get_contact_id(self) -> int:
        return self.__contact_id

    def get_user_id(self) -> int:
        return self.__user_id

    def get_first_name(self) -> str:
        return self.__first_name

    def is_email_address(self):
        return self.__telephone_number is not None

    def is_telephone_number(self):
        return self.__telephone_number is not None

    def get_email_address(self):
        return self.__email_address_str

    def get_telephone_address(self):
        return self.__telephone_number is not None

    def get_preferred_lang_code_str(self):
        return self.__preferred_lang_code_str

    # TODO Is there a package/library for this that we can use?
    def get_normizalied_phone(self, region: str = "IL"):
        """ normalized/canonical phone, telephone number """
        # TODO I think this function does not cover all cases - I believe we can find one on the internet - I think Sahar found one.
        if self.__telephone_number is None:
            return None
        return PhonesLocal.normalize_phone_number(self.__telephone_number, region)['full_number_normalized']


    def to_json(self):
        init_args = ("contact_id", "person_id", "user_id", "profile_id", "telephone_number",
                     "email_address_str", "preferred_lang_code_str", "first_name")
        return {k.replace("_Recipient__", ""): v for k, v in self.__dict__.items()
                if v is not None and k.replace("_Recipient__", "") in init_args}

    @staticmethod
    def from_json(json_object: dict) -> 'Recipient':
        return Recipient(**json_object)

    def __str__(self):
        return json.dumps(self.to_json())

    def __repr__(self):
        """This is used when we print a list of recipients"""
        return self.__str__()

    @staticmethod
    def recipients_from_json(recipients: list[dict]) -> list['Recipient'] | list | None:
        """get recipients"""
        if not recipients:
            return recipients  # None or []
        return [Recipient.from_json(recipient) for recipient in recipients]

    @staticmethod
    def recipients_to_json(recipients: list['Recipient']) -> list[dict] | list | None:
        """recipients to json"""
        if not recipients:
            return recipients
        return [recipient.to_json() for recipient in recipients]



import setuptools

PACKAGE_NAME = "message-local"
package_dir = PACKAGE_NAME.replace("-", "_")

setuptools.setup(
    name=PACKAGE_NAME,
    version='0.0.110', # https://pypi.org/project/message-local
    author="Circles",
    author_email="info@circles.life",
    url=f"https://github.com/circles-zone/message-local-python-package",
    packages=[package_dir],
    package_dir={package_dir: f'{package_dir}/src'},
    package_data={package_dir: ['*.py']},
    long_description="message-local",
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    install_requires=["item-local>=0.0.4",  # https://pypi.org/project/item-local/
                      "api-management-local>=0.0.39",  # https://pypi.org/project/api-management-local/
                      "language-local",
                      "variable-local",
                      "logger-local",
                      "database-mysql-local>=0.0.199",
                      "star-local",
                      "profiles-local>=0.0.50",
                      "phones-local"
                      ]
)



# TODO Please make sure every src file will have a test file i.e. CompoundMessage

import json
import random

from database_mysql_local.generic_crud import GenericCRUD
from logger_local.MetaLogger import MetaLogger
from profiles_local.profiles_local import ProfilesLocal
from user_context_remote.user_context import UserContext
from variable_local.template import ReplaceFieldsWithValues

from .MessageChannels import MessageChannel
from .MessageConstants import object_message
from .MessageTemplates import MessageTemplates
from .Recipient import Recipient

lang_code_cache = {}
JSON_VERSION = "240405"  # TODO: make it generic


class CompoundMessage(GenericCRUD, metaclass=MetaLogger, object=object_message):
    def __init__(self, campaign_id: int = None, message_template_id: int = None, body: str = None, subject: str = None,
                 recipients: list[Recipient] = None, message_id: int = None, is_test_data: bool = False, ui_schema=None,
                 schema=None, field_id=None, form_id: int = None):
        self.user_context = UserContext()
        super().__init__(default_schema_name="message", default_table_name="message_table",
                         default_view_table_name="message_sent_view", default_id_column_name="message_id",
                         is_test_data=is_test_data)

        self.campaign_id = campaign_id
        self.message_template_id = message_template_id
        self.body = body
        self.subject = subject
        self.recipients = recipients
        self.message_id = message_id
        self.form_id = form_id
        self.__compound_message = {}
        self.ui_schema = ui_schema
        self.schema = schema
        self.field_id = field_id
        self.profile_local = ProfilesLocal()
        self.message_template = MessageTemplates()
        self.set_compound_message_after_text_template()

    def _get_template_ids_by_campaign(self, campaign_id: int) -> list[int]:
        """Returns a random template id from the campaign"""

        possible_template_ids = self.select_multi_tuple_by_id(
            schema_name="campaign", view_table_name="campaign_view",
            id_column_name="campaign_id", id_column_value=campaign_id,
            select_clause_value="message_template_id")

        return [message_template_id[0] for message_template_id in possible_template_ids]

    # TODO: {
    #   "json_version": ...,
    #   "data": {
    #     "formId": ...,
    #     "formName": ...,
    #     "messageTemplates": [
    #     {
    #         "messageTemplateId": ...,
    #         "compoundMessages": [...]
    def set_compound_message_after_text_template(
            self, campaign_id: int = None, message_template_id: int = None, body: str = None, subject: str = None,
            recipients: list[Recipient] = None, message_id: int = None, form_id: int = None) -> None:
        """ Sets the compound message in the instance and in the database using the following structure:"""

        #     {
        #         # "compoundMessageId": ...,
        #         "DEFAULT": {
        #             "bodyBlocks": [
        #                 {
        #                     "messageTemplateTextBlockId": ...,
        #                     "blockTypeId": ...,
        #                     "messageTemplateId": ...,
        #                     "messageTemplateBlockTypeName": ...,
        #                     "questionTitle": ...,
        #                     "questionTypeName": ...,
        #                     "questionId": ...,
        #                     "questionTypeId": ...,
        #                     "answerExpirationDays": ...,
        #                     "tableId": ...,
        #                     "variableId": ...,
        #                     "whereStatement": ...,
        #                     "isVisible": ...,
        #                     "isRequired": ...,
        #                     "defaultValue": ...,
        #                     "labelQuestionId": ...,
        #                     "template": ...,
        #                     "hint": ...,
        #                     "profileBlocks": [
        #                         {"profileId": ..., "processedTemplate": ...},
        #                         {"profileId": ..., "processedTemplate": ...},
        #                     ]
        #                 },
        #                 {"messageTemplateTextBlockId": ...}
        #             ],
        #             "subjectBlock": {"messageTemplateTextBlockId": ...}  # same structure as one bodyBlock
        #         },
        #         "EMAIL": {"bodyBlocks": [...], "subjectBlock": {...}},  # same structure as DEFAULT
        #         "SMS": {},
        #         "WHATSAPP": {},
        #         ...
        #     }

        # Allow overiding instance vars
        campaign_id = campaign_id or self.campaign_id
        message_template_id = message_template_id or self.message_template_id
        body = body or self.body
        subject = subject or self.subject
        recipients = recipients or self.recipients or []
        message_id = message_id or self.message_id
        form_id = form_id or self.form_id

        data = {"DEFAULT": {"bodyBlocks": [], "subjectBlock": {}, "formId": form_id}}

        channels_mapping = {
            MessageChannel.EMAIL.name: {"body": "email_body_html_template", "subject": "email_subject_template"},
            "DEFAULT": {"body": "default_body_template", "subject": "default_subject_template"},
        }
        for channel in MessageChannel:
            data[channel.name] = {"bodyBlocks": [], "subjectBlock": {}, "formId": form_id}
            if channel.name not in channels_mapping:
                channels_mapping[channel.name] = {"body": f"{channel.name.lower()}_body_template",
                                                  "subject": f"{channel.name.lower()}_subject_template"}

        if body:
            textblocks_and_attributes = [{}]  # one textblock
            for message_channel, template_header in channels_mapping.items():
                textblocks_and_attributes[0][template_header["body"]] = body
                textblocks_and_attributes[0][template_header["subject"]] = subject

        else:  # If body is not given, get it from the database
            # if form_id:
            #     textblocks_and_attributes = self.message_template.get_textblocks_and_attributes_by_form_id(form_id=form_id)
            # else:
            textblocks_and_attributes = self.message_template.get_textblocks_and_attributes()

        if campaign_id and not message_template_id:
            possible_template_ids = self._get_template_ids_by_campaign(campaign_id=campaign_id)
            message_template_id = random.choice(possible_template_ids)
            self.logger.info(object={"message_template_id": message_template_id})

        criteria_json = self.message_template.get_critiria_json(message_template_id=message_template_id)
        potentials_recipients = self.message_template.get_potentials_recipients(
            criteria_json, recipients)
        self.logger.info(object={"textblocks_and_attributes": textblocks_and_attributes,
                                 "criteria_json": criteria_json})

        for message_template_textblock_and_attributes in textblocks_and_attributes:
            for message_channel, template_header in channels_mapping.items():
                for part in ("body", "subject"):
                    templates = [x for x in (message_template_textblock_and_attributes.get(template_header[part]),
                                             message_template_textblock_and_attributes.get("questionTitle"))
                                 if x]

                    message_template = " ".join(templates)
                    if not message_template:
                        self.logger.warning("message_template is empty", object={
                            "message_template_textblock_and_attributes": message_template_textblock_and_attributes})
                        continue

                    block = {
                        "messageTemplateTextBlockId": message_template_textblock_and_attributes.get("messageTemplateTextBlockId"),
                        "blockTypeId": message_template_textblock_and_attributes.get("blockTypeId"),
                        "messageTemplateId": message_template_textblock_and_attributes.get("messageTemplateId"),
                        "messageTemplateBlockTypeName": message_template_textblock_and_attributes.get(
                            "messageTemplateBlockTypeName"),
                        "template": message_template,
                        "questionTitle": message_template_textblock_and_attributes.get("questionTitle"),
                        "questionTypeName": message_template_textblock_and_attributes.get("questionTypeName"),
                        "schemaDefault": message_template_textblock_and_attributes.get("schemaDefault"),
                        "uiSchemaDefault": message_template_textblock_and_attributes.get("uiSchemaDefault"),
                        "schemaAttributes": message_template_textblock_and_attributes.get("schemaAttributes"),
                        "uiSchemaAttributes": message_template_textblock_and_attributes.get("uiSchemaAttributes"),
                        "messageTemplateName": message_template_textblock_and_attributes.get("messageTemplateName"),
                        "messageTemplateTextBlockSeq": message_template_textblock_and_attributes.get(
                            "messageTemplateTextBlockSeq"),
                        "variableName": message_template_textblock_and_attributes.get("variableName"),
                        "variableMlTitle": message_template_textblock_and_attributes.get("variableMlTitle"),
                        "fieldName": message_template_textblock_and_attributes.get("fieldName"),

                        # from question_view:
                        "questionId": message_template_textblock_and_attributes.get("questionId"),
                        "questionTypeId": message_template_textblock_and_attributes.get("questionTypeId"),
                        "possibleAnswers": message_template_textblock_and_attributes.get("possibleAnswers"),
                        "answerExpirationDays": message_template_textblock_and_attributes.get("answerExpirationDays"),
                        "tableId": message_template_textblock_and_attributes.get("tableId"),
                        "variableId": message_template_textblock_and_attributes.get("variableId"),
                        "whereStatement": message_template_textblock_and_attributes.get("whereStatement"),
                        "isVisible": message_template_textblock_and_attributes.get("isVisible"),
                        "isRequired": message_template_textblock_and_attributes.get("isRequired"),
                        "defaultValue": message_template_textblock_and_attributes.get("defaultValue"),
                        "labelQuestionId": message_template_textblock_and_attributes.get("labelQuestionId"),
                        "hint": message_template_textblock_and_attributes.get("hint"),
                        "profileBlocks": []
                    }

                    for recipient in recipients:
                        if recipient.get_profile_id() not in lang_code_cache:
                            lang_code_cache[
                                recipient.get_profile_id()] = self.profile_local.get_preferred_lang_code_by_profile_id(
                                recipient.get_profile_id()).value
                        # TODO: The test critiria doesn't match the recipient
                        if 1 or (any(recipient.get_profile_id() == potential_recipient["profile_id"]
                                     for potential_recipient in potentials_recipients)
                                 # TODO recipient_preferred_lang_code_str
                                 and (lang_code_cache[recipient.get_profile_id()] ==
                                      message_template_textblock_and_attributes.get("langCode"))):
                            block["profileBlocks"].append(
                                # each profile has its own template, because of the language
                                {"profileId": recipient.get_profile_id(),
                                 "processedTemplate": self._process_text_block(message_template, recipient=recipient),
                                 })
                    if part == "subject":
                        # ignore if already set
                        data[message_channel]["subjectBlock"] = data[message_channel]["subjectBlock"] or block
                    else:
                        data[message_channel][f"{part}Blocks"].append(block)

        # remove empty channels:
        data = {k: v for k, v in data.items() if v["bodyBlocks"] or v["subjectBlock"]}

        compound_message = {"json_version": JSON_VERSION, "data": data}
        # save in message table
        if message_id:
            self.update_by_id(id_column_value=message_id,
                              data_json={"compound_message_json": json.dumps(compound_message),
                                         "compound_message": json.dumps(compound_message),
                                         "compound_message_json_version": JSON_VERSION})
        else:
            self.message_id = self.insert(data_json={"compound_message_json": json.dumps(compound_message),
                                                     "compound_message": json.dumps(compound_message),
                                                     "compound_message_json_version": JSON_VERSION})

        self.__compound_message = compound_message
        self.logger.debug(object=locals())

    def _process_text_block(self, text_block_body: str, recipient: Recipient) -> str:
        template = ReplaceFieldsWithValues(message=text_block_body,
                                           # TODO language -> lang_code_str ?
                                           language=recipient.get_preferred_lang_code_str(),
                                           variable=recipient.get_profile_variables())
        processed_text_block = template.get_variable_values_and_chosen_option(
            profile_id=self.user_context.get_effective_profile_id(),
            # TODO profile_id -> effective_profile_id
            kwargs={"recipient.first_name": recipient.get_first_name(),
                    # TODO: effective -> real in user_context
                    "sender.first_name": self.user_context.get_real_first_name(),
                    "message_id": self.message_id
                    })
        return processed_text_block

    def get_compound_message_dict(self, channel: MessageChannel = None) -> dict:
        if channel is None:
            return self.__compound_message["data"]
        else:
            return {"DEFAULT": self.__compound_message["data"]["DEFAULT"],
                    channel.name: self.__compound_message["data"].get(channel.name, {})}

    def get_compound_message_str(self, channel: MessageChannel = None) -> str:
        return json.dumps(self.get_compound_message_dict(channel=channel))

    def get_profile_block(self, profile_id: int, channel: MessageChannel, part: str = "body") -> dict:
        """Returns a dict with the following keys:
        profileId, template, processedTemplate, messageTemplateTextBlockId, blockTypeId, messageTemplateBlockTypeName, questionId,
            questionTypeId, questionTitle, questionTypeName
        """

        assert part in ("body", "subject")
        if part == "body":
            blocks = self.__compound_message["data"].get(channel.name, {}).get("bodyBlocks", [])
        else:
            blocks = [self.__compound_message["data"].get(channel.name, {}).get("subjectBlock", {})]
        for block in blocks:
            for _profile_block in block.get("profileBlocks", []):
                if _profile_block.get("profileId") == profile_id:
                    return {**_profile_block, **{k: v for k, v in block.items() if k != "profileBlocks"}}
        return {}

    def get_message_fields(self) -> dict:
        if self.recipients:
            recipients_mapping = {recipient.get_profile_id(): recipient.to_json() for recipient in self.recipients}
        else:
            recipients_mapping = {}
        return {
            "campaign_id": self.campaign_id,
            "body": self.body,
            "subject": self.subject,
            "message_id": self.message_id,
            "ui_schema": self.ui_schema,
            "schema": self.schema,
            "field_id": self.field_id,
            "recipients": recipients_mapping,
            "json_version": JSON_VERSION,
        }



from enum import Enum


# TODO Let's generate with Sql2Code
# Values are from the database message_channel_table
class MessageChannel(Enum):
    EMAIL = 1
    SMS = 2
    MMS = 3
    RCS = 4
    FACEBOOK = 5
    TWITTER = 6
    LINKEDIN = 7
    SKYPE = 8
    TELEGRAM = 9
    DISCORD = 10
    WHATSAPP = 11
    WECHAT = 12
